
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

extern "C" {
#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
}

#include "php_array2arraymap_sdb.h"

/* If you declare any globals in php_array2arraymap_sdb.h uncomment this:
ZEND_DECLARE_MODULE_GLOBALS(array2arraymap_sdb)
*/

/* True global resources - no need for thread safety here */
static int le_array2arraymap_sdb;

/* {{{ array2arraymap_sdb_functions[]
 *
 * Every user visible function must have an entry in array2arraymap_sdb_functions[].
 */
zend_function_entry array2arraymap_sdb_functions[] = {
	PHP_FE(array2arraymap_sdb_open,	NULL)
	PHP_FE(array2arraymap_sdb_type,	NULL)
	PHP_FE(array2arraymap_sdb_get,	NULL)
	PHP_FE(array2arraymap_sdb_find,	NULL)
	PHP_FE(array2arraymap_sdb_find_pair,	NULL)
	PHP_FE(array2arraymap_sdb_find_pair_by_prefix,	NULL)
	PHP_FE(array2arraymap_sdb_exists_at,	NULL)
	PHP_FE(confirm_array2arraymap_sdb_compiled,	NULL)		/* For testing, remove later. */
	{NULL, NULL, NULL}	/* Must be the last line in array2arraymap_sdb_functions[] */
};
/* }}} */

/* {{{ array2arraymap_sdb_module_entry
 */
zend_module_entry array2arraymap_sdb_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	"array2arraymap_sdb",
	array2arraymap_sdb_functions,
	PHP_MINIT(array2arraymap_sdb),
	PHP_MSHUTDOWN(array2arraymap_sdb),
	NULL, /* PHP_RINIT(array2arraymap_sdb),		/* Replace with NULL if there's nothing to do at request start */
	NULL, /* PHP_RSHUTDOWN(array2arraymap_sdb),	/* Replace with NULL if there's nothing to do at request end */
	PHP_MINFO(array2arraymap_sdb),
#if ZEND_MODULE_API_NO >= 20010901
	"0.1", /* Replace with version number for your extension */
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_ARRAY2ARRAYMAP_SDB
BEGIN_EXTERN_C()
ZEND_GET_MODULE(array2arraymap_sdb)
END_EXTERN_C()
#endif

/* {{{ PHP_INI
 */
/* Remove comments and fill if you need to have entries in php.ini
PHP_INI_BEGIN()
    STD_PHP_INI_ENTRY("array2arraymap_sdb.global_value",      "42", PHP_INI_ALL, OnUpdateLong, global_value, zend_array2arraymap_sdb_globals, array2arraymap_sdb_globals)
    STD_PHP_INI_ENTRY("array2arraymap_sdb.global_string", "foobar", PHP_INI_ALL, OnUpdateString, global_string, zend_array2arraymap_sdb_globals, array2arraymap_sdb_globals)
PHP_INI_END()
*/
/* }}} */

/* {{{ php_array2arraymap_sdb_init_globals
 */
/* Uncomment this function if you have INI entries
static void php_array2arraymap_sdb_init_globals(zend_array2arraymap_sdb_globals *array2arraymap_sdb_globals)
{
	array2arraymap_sdb_globals->global_value = 0;
	array2arraymap_sdb_globals->global_string = NULL;
}
*/
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(array2arraymap_sdb)
{
	/* uncomment this line if you have INI entries
	UNREGISTER_INI_ENTRIES();
	*/
	return SUCCESS;
}
/* }}} */

/* Remove if there's nothing to do at request start */
/* {{{ PHP_RINIT_FUNCTION
 */
PHP_RINIT_FUNCTION(array2arraymap_sdb)
{
	return SUCCESS;
}
/* }}} */

/* Remove if there's nothing to do at request end */
/* {{{ PHP_RSHUTDOWN_FUNCTION
 */
PHP_RSHUTDOWN_FUNCTION(array2arraymap_sdb)
{
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(array2arraymap_sdb)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "array2arraymap_sdb support", "enabled");
	php_info_print_table_end();

	/* Remove comments if you have entries in php.ini
	DISPLAY_INI_ENTRIES();
	*/
}
/* }}} */


/*---------------------------------------------------*/
#include <mapsdb/SDB.h>

/* Resource struct, not the object :) */
/* Extends zend object */
typedef struct _ze_sdb_object {
	zend_object zo;
	SDB *sdb_ptr;
} ze_sdb_object;

// hack
#define sdb_object SDB

static zend_class_entry *sdb_class_entry_ce;

static void sdb_free_resource_ptr(sdb_object *intern TSRMLS_DC);

/* {{{ sdb_object_free_storage */
static void sdb_free_resource_ptr(sdb_object *intern TSRMLS_DC) 
{
	if (intern) {
		delete(intern);
	}
}
/* }}} */

#ifdef ZEND_ENGINE_2
/* {{{ SDB_FROM_OBJECT */
#define SDB_FROM_OBJECT(intern, object) \
	{ \
		ze_sdb_object *obj = (ze_sdb_object*) zend_object_store_get_object(object TSRMLS_CC); \
		intern = obj->sdb_ptr; \
		if (!intern) { \
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid or unitialized SDB object"); \
			RETURN_FALSE; \
		} \
	}
/* }}} */

static zend_object_handlers sdb_object_handlers;

/* {{{ sdb_object_free_storage */
static void sdb_object_free_storage(void *object TSRMLS_DC)
{
	ze_sdb_object * intern = (ze_sdb_object *) object;
	if (!intern) {
		return;
	}
	if (intern->sdb_ptr) {
		sdb_free_resource_ptr(intern->sdb_ptr TSRMLS_CC);
	}
	intern->sdb_ptr = NULL;
	zend_object_std_dtor(&intern->zo TSRMLS_CC);
	
	efree(intern);
}
/* }}} */


/* {{{ sdb_object_new */
static zend_object_value sdb_object_new(zend_class_entry *class_type TSRMLS_DC)
{
	ze_sdb_object *intern;
	zval *tmp;
	zend_object_value retval;

	intern = (ze_sdb_object*) emalloc(sizeof(ze_sdb_object));
	memset(&intern->zo, 0, sizeof(zend_object));
	intern->sdb_ptr = NULL;
	
	zend_object_std_init(&intern->zo, class_type TSRMLS_CC);
#if PHP_VERSION_ID < 50399	
	zend_hash_copy(intern->zo.properties, &class_type->default_properties, (copy_ctor_func_t) zval_add_ref,
					(void *) &tmp, sizeof(zval *));
#else
	object_properties_init(&(intern->zo), class_type);
#endif

	retval.handle = zend_objects_store_put(intern,
						NULL,
						(zend_objects_free_object_storage_t) sdb_object_free_storage,
						NULL TSRMLS_CC);
	
	retval.handlers = (zend_object_handlers *) & sdb_object_handlers;
	
	return retval;
}
/* }}} */
#endif


#ifdef ZEND_ENGINE_2
/* {{{ sdb_class_functions */
static zend_function_entry array2arraymap_sdb_class_functions[] = {
	PHP_ME_MAPPING(open,	array2arraymap_sdb_open, 	NULL, 0)
	PHP_ME_MAPPING(type,	array2arraymap_sdb_type, 	NULL, 0)
	PHP_ME_MAPPING(get,	array2arraymap_sdb_get, 	NULL, 0)
	PHP_ME_MAPPING(find,	array2arraymap_sdb_find, 	NULL, 0)
	PHP_ME_MAPPING(find_pair,	array2arraymap_sdb_find_pair, 	NULL, 0)
	PHP_ME_MAPPING(find_pair_by_prefix,	array2arraymap_sdb_find_pair_by_prefix, 	NULL, 0)
	PHP_ME_MAPPING(exists_at,	array2arraymap_sdb_exists_at, 	NULL, 0)
	{NULL, NULL, NULL}
};
/* }}} */
#endif


/* {{{ proto resource array2arraymap_sdb_open()
Create new array2arraymap_sdb from file */
static PHP_FUNCTION(array2arraymap_sdb_open)
{
	sdb_object *intern;

	zval *php_this = getThis();
	ze_sdb_object *ze_obj = NULL;

	char *name;
    int name_len;

	if (php_this) {
		/* We do not use SDB_FROM_OBJECT, sdb init function here */
		ze_obj = (ze_sdb_object*) zend_object_store_get_object(php_this TSRMLS_CC); 
	} else RETURN_FALSE;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &name, &name_len) == FAILURE) {
        RETURN_NULL();
    }

    //php_printf("Hello %s ", name);

	//intern = emalloc(sizeof(sdb_object));
	try {
		intern = new SDB(name);
		if (php_this) {
			ze_obj->sdb_ptr = intern;
			RETURN_TRUE;
		}
	} catch (sdb_error e) {
		// todo: some error messaging here
		std::cerr << e.what() << '\n';
	}
	RETURN_FALSE;
}
/* }}} */

// C++ hacks
//define Z_LVAL_P(zval)	(*zval).value.lval


/* can't inline; require ZEND_FUNCTION expanded */
/*
inline
void free_path(zval** parameter_array[]) {
	if (parameter_array!=NULL) efree(parameter_array);
}
inline
zval** *get_parameter_array(long argument_count) {
	zval ** *parameter_array = (zval***) ecalloc(argument_count, sizeof(zval **));
	if(zend_get_parameters_array_ex(argument_count, parameter_array) != SUCCESS)
		return NULL;
	return parameter_array;
}
*/

/* can't inline; require ZEND_FUNCTION expanded */
/* proto: zval** get_path(sdb_object* intern); */
#define get_path(intern) \
	zval ** *parameter_array; \
	zval ** cur_zval; \
	/* get the number of arguments */ \
	long argument_count = ZEND_NUM_ARGS(); \
	if(argument_count == 0) WRONG_PARAM_COUNT; \
	\
	/* argument count is correct, now retrieve arguments */ \
	parameter_array = (zval***) ecalloc(argument_count, sizeof(zval **)); \
	if(zend_get_parameters_array_ex(argument_count, parameter_array) != SUCCESS) \
		WRONG_PARAM_COUNT; \
	if (parameter_array==NULL) WRONG_PARAM_COUNT; \
	intern->reset(); \
	if (IS_ARRAY==Z_TYPE_PP(parameter_array[0])) { \
		HashPosition pointer; \
		HashTable *arr_hash = Z_ARRVAL_PP(parameter_array[0]); \
		long array_count = zend_hash_num_elements(arr_hash); \
		for(zend_hash_internal_pointer_reset_ex(arr_hash, &pointer); (zend_hash_get_current_data_ex(arr_hash, (void**) &cur_zval, &pointer) == SUCCESS) && (array_count > 1); zend_hash_move_forward_ex(arr_hash, &pointer)) { \
			convert_to_long_ex(cur_zval);				\
			intern->set_next(Z_LVAL_P(*cur_zval));		\
			array_count--; \
        } \
		zend_hash_get_current_data_ex(arr_hash, (void**) &cur_zval, &pointer); \
    } else { \
		for (int i = 0; i < argument_count-1; i++) {	\
			cur_zval = parameter_array[i];				\
			convert_to_long_ex(cur_zval);				\
			intern->set_next(Z_LVAL_P(*cur_zval));		\
		} \
		cur_zval = parameter_array[argument_count-1]; \
	} \
// } //return cur_zval;


typedef void (ZVAL_CREATOR_NUM)(zval* array, large_int num, const char* top);

void creator_iv (zval* array, large_int i, const char* top) {
	add_next_index_long(array, i);
}
void creator_pv (zval* array, large_int i, const char* top) {
	const char *str = top + i;
	add_next_index_string(array, (char*) str, 1);
}

inline 
void wrap_val(table < noid_type,noid_type >& tbl, std::pair<num_iter, num_iter> p, zval *results, ZVAL_CREATOR_NUM key_creator,ZVAL_CREATOR_NUM val_creator, bool pair = false) {
  const char* top = tbl.get_data_offset();
  for (num_iter cur = p.first; cur !=p.second; ++cur) {
    if(pair) key_creator(results, cur.get_key(), top);
    val_creator(results, cur.get_val(), top);
  }
}

inline
void map_int(table < noid_type,noid_type >& tbl,  large_int n, zval *results, ZVAL_CREATOR_NUM key_creator,ZVAL_CREATOR_NUM val_creator, bool pair = false) {
  wrap_val(tbl, num_iter::map_numeric_key(&tbl, n), results, key_creator, val_creator, pair);
}

inline
void map_str(table < noid_type,noid_type >& tbl, char* str, zval *results, ZVAL_CREATOR_NUM key_creator,ZVAL_CREATOR_NUM val_creator, bool pair = false, NodeConst::comparison_type tt = NodeConst::FULL) {
  wrap_val(tbl, num_iter::map_string_key(&tbl, str, tt), results, key_creator, val_creator, pair);
}

ZVAL_CREATOR_NUM* choose_creator(NodeConst::TYPE type) {
  switch (type) {
  case NodeConst::INT4:
  case NodeConst::INT8:
     return &creator_iv; break;
  case NodeConst::STRING4:
    return &creator_pv; break;
  case NodeConst::CONTEXT4:
  case NodeConst::CONTEXT8:
    throw sdb_error_incomplete_path();
  default:
    fprintf(stderr, "can't find value creator for Type %d !\n", type);
    throw sdb_error("can't find value creator for Type");
  }
  return &creator_iv;
}

large_int convert_sv_to_num (zval* val, NodeConst::TYPE type) {
  switch (type) {
  case NodeConst::INT4:
  case NodeConst::INT8:
    return Z_LVAL_P(val); break;
  case NodeConst::STRING4:
  default:
    fprintf(stderr, "internal error: unknown type %d for convert_sv_to_num !\n", type);
    throw sdb_error("internal error: unknown type for convert_sv_to_num !");
  }
  return 0;
}

void map_common (table < noid_type,noid_type > tbl, zval* key, zval *results, bool is_pair = false, NodeConst::comparison_type tt = NodeConst::PREFIX) {
  NodeConst::TYPE keytype = tbl.get_keytype();
  NodeConst::TYPE valtype = tbl.get_valtype();
  ZVAL_CREATOR_NUM* key_creator = choose_creator(keytype);
  ZVAL_CREATOR_NUM* val_creator = choose_creator(valtype);
  switch (keytype) {
  case NodeConst::INT4:
  case NodeConst::INT8:
    map_int(tbl, convert_sv_to_num(key, keytype), results, *key_creator, *val_creator, is_pair);
    break;
  case NodeConst::STRING4:
    map_str(tbl, Z_STRVAL_P(key), results, *key_creator, *val_creator, is_pair, tt);
    break;
  default:
    fprintf(stderr, "Type %d is not supported in map_common!\n", keytype);
  }
}

void find_common (SDB *THIS, zval* key, zval *results, bool is_pair = false, NodeConst::comparison_type tt = NodeConst::FULL) {
  map_common (THIS->current()->get_find_map(), key, results, is_pair, tt);
}

/* ######################################################################### */


/* {{{ proto string sdb_type(SDB sdb, path[])
generic SDB func template */
static PHP_FUNCTION(array2arraymap_sdb_type)
{
	sdb_object *intern;
	int retval;
	zval *php_this = getThis();
	
	if (php_this) {
		SDB_FROM_OBJECT(intern, php_this);
	}
	if (intern) {
		try {
			get_path(intern);
			intern->set_next(Z_LVAL_P(*cur_zval));
			RETURN_STRING((char*) NodeConst::get_type_name(intern->current()->get_valtype()), 1);
		} catch (sdb_error_no_path e) {
			RETURN_FALSE;
		}
	}
	RETURN_FALSE;
}
/* }}} */

/* {{{ proto array sdb_bla_bla(SDB sdb, path[])
generic SDB func template */
static PHP_FUNCTION(array2arraymap_sdb_get)
{
	sdb_object *intern;
	int retval;
	zval *php_this = getThis();
	
	if (php_this) {
		SDB_FROM_OBJECT(intern, php_this);
	}
	if (intern) {
		array_init(return_value);
		try {
			get_path(intern);
			map_common (intern->current()->get_get_map(), *cur_zval, return_value);
		} catch (sdb_error_no_path e) {
		} catch (sdb_error_incomplete_path e) {
			std::cerr << e.what() << '\n';
		}
	}
}
/* }}} */

/* {{{ proto array sdb_bla_bla(SDB sdb, path[])
generic SDB func template */
static PHP_FUNCTION(array2arraymap_sdb_find)
{
	sdb_object *intern;
	int retval;
	zval *php_this = getThis();
	
	if (php_this) {
		SDB_FROM_OBJECT(intern, php_this);
	}
	if (intern) {
		array_init(return_value);
		try {
			get_path(intern);
			find_common (intern, *cur_zval, return_value, false);
		} catch (sdb_error_no_path e) {
		} catch (sdb_error_incomplete_path e) {
			std::cerr << e.what() << '\n';
		}
	}
}
/* }}} */

/* {{{ proto array sdb_bla_bla(SDB sdb, path[])
generic SDB func template */
static PHP_FUNCTION(array2arraymap_sdb_find_pair)
{
	sdb_object *intern;
	int retval;
	zval *php_this = getThis();
	
	if (php_this) {
		SDB_FROM_OBJECT(intern, php_this);
	}
	if (intern) {
		array_init(return_value);
		try {
			get_path(intern);
			find_common (intern, *cur_zval, return_value, true);
		} catch (sdb_error_no_path e) {
		} catch (sdb_error_incomplete_path e) {
			std::cerr << e.what() << '\n';
		}
	}
}
/* }}} */

/* {{{ proto array sdb_bla_bla(SDB sdb, path[])
generic SDB func template */
static PHP_FUNCTION(array2arraymap_sdb_find_pair_by_prefix)
{
	sdb_object *intern;
	int retval;
	zval *php_this = getThis();
	
	if (php_this) {
		SDB_FROM_OBJECT(intern, php_this);
	}
	if (intern) {
		array_init(return_value);
		try {
			get_path(intern);
			find_common (intern, *cur_zval, return_value, true, NodeConst::PREFIX);
		} catch (sdb_error_no_path e) {
		} catch (sdb_error_incomplete_path e) {
			std::cerr << e.what() << '\n';
		}
	}
}
/* }}} */

/* {{{ proto array sdb_bla_bla(SDB sdb, path[])
generic SDB func template */
static PHP_FUNCTION(array2arraymap_sdb_exists_at)
{
	sdb_object *intern;
	int retval;
	zval *php_this = getThis();
	long noid = 0;
	int argc = ZEND_NUM_ARGS( ); 
    if (zend_parse_parameters(argc TSRMLS_CC, "l", &noid) == FAILURE) 
        RETURN_FALSE;	
	if (php_this) {
		SDB_FROM_OBJECT(intern, php_this);
	}
	if (intern) {
		intern->reset();
		if (intern->exists_at(noid)) RETURN_TRUE
		else RETURN_FALSE;
	}
}
/* }}} */


















/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(array2arraymap_sdb)
{
	zend_class_entry ce;
	memcpy(&sdb_object_handlers, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
	sdb_object_handlers.clone_obj      = NULL;
	INIT_CLASS_ENTRY(ce, "SDB", array2arraymap_sdb_class_functions);
	ce.create_object = sdb_object_new;
	sdb_class_entry_ce = zend_register_internal_class(&ce TSRMLS_CC);
	return SUCCESS;
}
/* }}} */

/*---------------------------------------------------*/













/* Remove the following function when you have succesfully modified config.m4
   so that your module can be compiled into PHP, it exists only for testing
   purposes. */

/* Every user-visible function in PHP should document itself in the source */
/* {{{ proto string confirm_array2arraymap_sdb_compiled(string arg)
   Return a string to confirm that the module is compiled in */
PHP_FUNCTION(confirm_array2arraymap_sdb_compiled)
{
	char *arg = NULL;
	int arg_len, len;
	char *strg;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &arg, &arg_len) == FAILURE) {
		return;
	}

	len = spprintf(&strg, 0, "Congratulations! You have successfully modified ext/%.78s/config.m4. Module %.78s is now compiled into PHP.", "array2arraymap_sdb", arg);
	RETURN_STRINGL(strg, len, 0);
}
/* }}} */
/* The previous line is meant for vim and emacs, so it can correctly fold and 
   unfold functions in source code. See the corresponding marks just before 
   function definition, where the functions purpose is also documented. Please 
   follow this convention for the convenience of others editing your code.
*/


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
